#include <algorithm>
#include <cassert>
#include <cstring>
#include <iostream>
#include <vector>
#include "answer.h"
#include "csmmm.h"
#include "questiongen.h"
#include "timer.h"

using namespace std;

/** Masets **/

class AllConfigurations : public Maset {
  public:
    AllConfigurations(int colors) {
      Peg peg_tbl[pegs];
      m.clear();
      Init(peg_tbl);
      m.push_back(Configuration(peg_tbl));
      while (Next(peg_tbl, colors))
        m.push_back(Configuration(peg_tbl));
    }

  private:
    void Init(Peg* peg_tbl) {
      for (int i = 0; i < pegs; ++i)
        peg_tbl[i] = 0;
    }

    bool Next(Peg* peg_tbl, int colors) {
      int i;
      for(i = 0; i < pegs; ++i) {
        if (peg_tbl[i] == colors - 1)
          peg_tbl[i] = 0;
        else {
          ++peg_tbl[i];
          break;
        }
      }
      return i < pegs;
    }
};

/** Asking questions and computing answers **/

class AnsweredMaset {
  public:
    AnsweredMaset(Maset const& mas, Configuration const& question) {
      asked_question = question;
      masets.resize(possible_answers);
      for (int i = 0; i < possible_answers; ++i)
        masets[i] = make_pair(Maset(), Answer(i));
      for (size_t i = 0; i < mas.size(); ++i)
        masets[Answer(mas[i], question).Code()].first.push(mas[i]);
      sort(masets.begin(), masets.end(), CompareMasets);
    }

    size_t size() const {
      return masets.size();
    }

    size_t max_size() const {
      return masets[0].first.size();
    }

    Maset const& GetMaset(size_t i) const {
      return masets[i].first;
    }

    Answer const& GetAnswer(size_t i) const {
      return masets[i].second;
    }

    Configuration const& GetAskedQuestion() const {
      return asked_question;
    }

    void swap(AnsweredMaset& a) {
      masets.swap(a.masets);
      asked_question.swap(a.asked_question);
    }

  private:
    typedef pair<Maset, Answer> MasetWithAnswer;

    vector<MasetWithAnswer> masets;
    Configuration asked_question;

    // Masets are sorted in the decreasing order of their size.
    static bool CompareMasets(MasetWithAnswer const &p1, MasetWithAnswer const &p2) {
      return p1.first.size() > p2.first.size();
    }
};

static void swap(AnsweredMaset& left, AnsweredMaset& right) {
  left.swap(right);
}

static ostream& operator<<(ostream& os, AnsweredMaset const& ans_mas) {
  os << "Present question " << ans_mas.GetAskedQuestion() << '\n';
  for (size_t i = 0; i < ans_mas.size(); ++i) {
    Maset const & mas = ans_mas.GetMaset(i);
    if (mas.size())
      os << "The answer "
         << ans_mas.GetAnswer(i)
         << " gives the maset\n"
         << mas;
  }
  // os << '\n';
  return os;
}

class AllAnswers {
  public:
    AllAnswers(Maset const& maset, Maset const& questions) {
      for (size_t i = 0; i < questions.size(); ++i) {
        AnsweredMaset ans_mas(maset, questions[i]);
        answered_masets.push_back(ans_mas);
      }
      sort(answered_masets.begin(), answered_masets.end(), CompareAnsweredMasets);
    }

    size_t size() const {
      return answered_masets.size();
    }

    AnsweredMaset const& GetAnsweredMaset(size_t i) const {
      return answered_masets[i];
    }

  private:
    vector<AnsweredMaset> answered_masets;

    // Answered masets are sorted in the increasing order of the size of the biggest maset.
    static bool CompareAnsweredMasets(AnsweredMaset const& p1, AnsweredMaset const& p2) {
      return p1.max_size() < p2.max_size();
    }
};

static ostream & operator<<(ostream& os, AllAnswers const& ans) {
  for (size_t i = 0; i < ans.size(); ++i) {
    os << ans.GetAnsweredMaset(i);
  }
  return os;
}

/** Solution search **/

static Maset const* pAllQuestions;
static Maset const* pAllSecrets;

class QuestionGeneratorWrapper {
  public:
    QuestionGeneratorWrapper() {
      #if BLACK_ONLY == 0 // white and black paegs in answers
      Maset prev_asked(questions);
      // Use nauty to exclude isomorphic questions.
      qm = NautyQuestionGenerator(*pAllQuestions, prev_asked);
      #else // black-peg only answers
      // There is only one non-isomorphic question.
      qm = new Maset({Configuration(Peg(0))});
      #endif
    }

    ~QuestionGeneratorWrapper() {
      delete qm;
    }

    Maset const & GetQuestions() {
      if (qm)
        return *qm;
      else
        return *pAllQuestions;
    }

  private:
    Maset const *qm;
};

typedef vector<Answer> AnswerVector;

static bool solvable(Maset  const& maset,
                     Answer const& answer,
                     Maset  const& question_set,
                     int          remaining_questions) {
  if (show) {
    cout << "maset\n" << maset << answer << ' ' << remaining_questions << '\n';
  }
  if (answer.B() == pegs) {
    if (show) {
      cout << "return true\n";
    }
    return true;
  }
  else if (remaining_questions == 0) {
    if (show) {
      cout << "return false\n";
    }
    return false;
  }
  remaining_questions--;
  AllAnswers ans(maset, question_set);
  if (show) {
    cout << ans << endl;
  }
  for (size_t i = 0; i < ans.size(); ++i) {
    bool res = false;
    for (size_t j = 0; j < ans.GetAnsweredMaset(i).size(); ++j) {
      if (ans.GetAnsweredMaset(i).GetMaset(j).size() > 0) {
        // GetMaset(j).size() ==  0 means that GetAnswer(j) is impossible for maset.
        res = solvable(ans.GetAnsweredMaset(i).GetMaset(j),
                       ans.GetAnsweredMaset(i).GetAnswer(j),
                       ans.GetAnsweredMaset(i).GetMaset(j),
                       remaining_questions);
        if (res == false) {
          break;
        }
      }
    }
    if (res) {
      return true;
    }
  }
  return false;
}

int main() {
  Timer t;

  cerr << "black pegs only = " << BLACK_ONLY << '\n'
       << "pegs            = " << pegs << '\n'
       << "colors          = " << colors << '\n'
       << "max questions   = " << questions << endl;

  pAllSecrets = new AllConfigurations(colors);
  pAllQuestions = pAllSecrets;

  QuestionGeneratorWrapper qgw;
  if (show) {
    cout << "the first question set:\n" << qgw.GetQuestions();
  }
  bool result = solvable(*pAllSecrets, Answer(0, 0), qgw.GetQuestions(), questions);

  cout << game_name << '0' << "(" << pegs << "," << colors << ") "
       << (result ? "<= " : "> ")
       << questions << '\n'
       << "computation time "
       << t.getTime()
       << '\n'
       << endl;

  return result;
}
