#ifndef _QUESTIONGEN_H_
#define _QUESTIONGEN_H_  1

#define USING_MAP 1

#if defined USING_MAP
  #include <map>
#elif defined USING_HASH
  #include <unordered_map>
#else
  #error Undefined storage container
#endif
#include "csmmm.h"

#define QUESTION_COLORS  COLORS

#define MAXN  (QUESTION_COLORS + PEGS * 2)
#include <nauty.h>

// Wrapper for nauty graph
class Graph {
  public:
    Graph(int vertices_number) {
      n = vertices_number;
      m = (n + WORDSIZE - 1) / WORDSIZE;
    }

    int getVerticesNumber() const {
      return n;
    }

    graph * getGraph() {
      return g;
    }

    bool operator==(Graph const& s) const {
      for (int i = 0; i < n * m; ++i)
        if (g[i] != s.g[i])
          return false;
      return true;
    }

    #if defined USING_MAP
    bool operator<(Graph const& s) const {
      for (int i = 0; i < n * m; ++i)
        if (g[i] < s.g[i])
          return true;
        else if (g[i] > s.g[i])
          return false;
      return false;
    }
    #endif

  protected:
    int n, m;
    graph g[MAXN*MAXM];

  #if defined USING_HASH
  friend struct ::std::hash<Graph>;
  #endif
};

#if defined USING_HASH
namespace std {
  template <>
  struct hash<Graph> {
    size_t operator()(Graph const& s) const {
      size_t h = 0;
      for (int i = 0; i < s.n * s.m; ++i)
        h += s.g[i];
      return h;
    }
  };
}
#endif

// graph representing a question sequence
class QGraph : public Graph {
  private:
    // tables used by nauty
    set* gv[MAXN];
    int lab[MAXN], ptn[MAXN], orbits[MAXN];

    void addHypherEdge(int x, int y, int z) {
      // A triple (x, y, z) represents a peg,
      // x is the column number, y is the row number and z is its color.
      // Add two edges {x+QUESTION_COLORS+PEGS+y*PEGS, x+QUESTION_COLORS},
      //               {x+QUESTION_COLORS+PEGS+y*PEGS, z}.
      ADDELEMENT(gv[x+QUESTION_COLORS+PEGS+y*PEGS], x+QUESTION_COLORS);
      ADDELEMENT(gv[x+QUESTION_COLORS+PEGS+y*PEGS], z);
      ADDELEMENT(gv[x+QUESTION_COLORS], x+QUESTION_COLORS+PEGS+y*PEGS);
      ADDELEMENT(gv[z],                 x+QUESTION_COLORS+PEGS+y*PEGS);
    }

    void delHypherEdge(int x, int y, int z) {
      // A triple (x, y, z) represents a peg,
      // x is the column number, y is the row number and z is its color.
      // Delete two edges {x+QUESTION_COLORS+PEGS+y*PEGS, x+QUESTION_COLORS},
      //                  {x+QUESTION_COLORS+PEGS+y*PEGS, z}.
      DELELEMENT(gv[x+QUESTION_COLORS+PEGS+y*PEGS], x+QUESTION_COLORS);
      DELELEMENT(gv[x+QUESTION_COLORS+PEGS+y*PEGS], z);
      DELELEMENT(gv[x+QUESTION_COLORS], x+QUESTION_COLORS+PEGS+y*PEGS);
      DELELEMENT(gv[z],                 x+QUESTION_COLORS+PEGS+y*PEGS);
    }

    void initLabPtn() {
      // Initialize tables used by nauty.
      for (int i = 0; i < n; ++i) {
        lab[i] = i;
        ptn[i] = 1;
      }
      #if COLORS != QUESTION_COLORS
      ptn[COLORS - 1] = 0; // Do not map the additional color to any other color.
      #endif
      for (int i = QUESTION_COLORS - 1; i < n; i += PEGS)
        ptn[i] = 0;
    }

  public:
    QGraph(int rows) : Graph(QUESTION_COLORS + PEGS + PEGS * rows) {
      // nauty_check(WORDSIZE, m, n, NAUTYVERSIONID);
      for (int i = 0; i < n; ++i) {
        gv[i] = GRAPHROW(g, i, m);
        EMPTYSET(gv[i], m);
      }
    }

    void addQuestion(int row, Configuration const &q) {
      for (int i = 0; i < PEGS; ++i)
        addHypherEdge(i, row, q[i]);
    }

    void delQuestion(int row, Configuration const &q) {
      for (int i = 0; i < PEGS; ++i)
        delHypherEdge(i, row, q[i]);
    }

    void callNauty(Graph *gc) {
      #define WRKSPC (400 * MAXM)
      static DEFAULTOPTIONS_GRAPH(options);
      static setword workspace[WRKSPC];
      static statsblk stats;
      options.getcanon = TRUE;
      options.defaultptn = FALSE;
      initLabPtn();
      nauty(g, lab, ptn, NULL, orbits, &options, &stats,
            workspace, WRKSPC, m, n, gc->getGraph());
      #undef WRKSPC
    }
};

static inline Maset const* NautyQuestionGenerator(Maset const& all_questions,
                                                  Maset const& asked_questions) {
  #if defined USING_MAP
  typedef ::std::map<Graph, Configuration> QSet;
  #elif defined USING_HASH
  typedef ::std::unordered_map<Graph, Configuration> QSet;
  #endif
  typedef ::std::pair<Graph, Configuration> QPair;

  int    level(asked_questions.size());
  QGraph q_graph(level + 1);
  Graph  canon_graph(q_graph.getVerticesNumber());
  #if defined USING_MAP
  QSet   question_set;
  #elif defined USING_HASH
  QSet   question_set(41507); // TODO: value
  #endif

  // Initialize graph which already asked questions.
  for (int i = 0; i < level; ++i)
    q_graph.addQuestion(i, asked_questions[i]);

  // Compute all non-isomorphic questions extending asked questions.
  for (size_t i = 0; i < all_questions.size(); ++i) {
    q_graph.addQuestion(level, all_questions[i]);
    q_graph.callNauty(&canon_graph);
    question_set.insert(QPair(canon_graph, all_questions[i]));
    q_graph.delQuestion(level, all_questions[i]);
  }

  // Do not repeat the same questions.
  for (int i = 0; i < level; ++i) {
    q_graph.addQuestion(level, asked_questions[i]);
    q_graph.callNauty(&canon_graph);
    question_set.erase(canon_graph);
    q_graph.delQuestion(level, asked_questions[i]);
  }

  // TODO: How to avoid copying?
  Maset *qm = new Maset(question_set.size());
  for (QSet::const_iterator it = question_set.begin();
       it != question_set.end();
       ++it) {
    qm->push(it->second);
  }

  return qm;
}

#endif
